/*

  WiFi Web Server with 6D IMU and Servo added to initially test IoT33

  This example is written for a network using WPA encryption. For
  WEP or WPA, change the Wifi.begin() call accordingly.

*/
#include <Servo.h>
#include <SPI.h>
#include <WiFiNINA.h>
#include "arduino_secrets.h"
#include <Arduino_LSM6DS3.h>
/*   This is a summary of Servo Library
  public:
  Servo();
  uint8_t attach(int pin);           // attach the given pin to the next free channel, sets pinMode, returns channel number or 0 if failure
  uint8_t attach(int pin, int min, int max); // as above but also sets min and max values for writes.
  void detach();
  void write(int value);             // if value is < 200 its treated as an angle, otherwise as pulse width in microseconds
  void writeMicroseconds(int value); // Write pulse width in microseconds
  int read();                        // returns current pulse width as an angle between 0 and 180 degrees
  int readMicroseconds();            // returns current pulse width in microseconds for this servo (was read_us() in first release)
  bool attached();                   // return true if this servo is attached, otherwise false
*/
#define VCC 3.305
#define LSB VCC/1024
float V0scale=(20+10)/10.0;
float V1scale=(20+10)/10.0;
float V2scale=(20+10)/10.0;
float V3scale=(39+10)/10.0;
// Analog read reference
//  Set the default physical values for each servo as mounted
#define ELSERVO_PIN 9
#define ELSERVO_MIN 905     // default settings for 0 (physical) degrees,  in us and varies somewhat with each servo
#define ELSERVO_MAX 2190    // "            180     

#define AZSERVO_PIN 10
#define AZSERVO_MIN  926
#define AZSERVO_MAX 2260

#define SERVO_HARDLIMIT_LOW 500   // absolute servo limtis
#define SERVO_HARDLIMIT_HIGH 2500
int i, micro_step = 7, slowdown = 10; //full range should take about 1.4 seconds at 1:1 (slowdown=1)scaling.
float  ElAngle = 0, AzAngle = 90, ElErrorAngle=0.0;  // User-referenced angles
// create  servo objects to control rotors
Servo azservo;
Servo elservo;
//
/////// enter  sensitive data in the Secret tab/arduino_secrets.h
char ssid[] = SECRET_SSID;        // your network SSID (name)
char pass[] = SECRET_PASS;    // your network password (use for WPA, or use as key for WEP)
int keyIndex = 0;                 // your network key Index number (needed only for WEP)

int status = WL_IDLE_STATUS;
int port = 8078;
WiFiServer server(port);


                // Considering the reference to be a vector pointing out the antenna mount axis, away from the Az/El assembly:
                // For user API calls, default setup of both rotors is mid-position with Az East=90 degrees and El H=0 degrees
                // Note that these positions are not "OEM Degrees" which go from 0 to 180 as defined by constants.
                // ElAngle is -90 to 90 with 0 up so offset -90 from OEM, AzAngle is 0-180 but rotates "backwards" compared to OEM 
                
boolean stepping = false;
int microsecs, prev_microsecs;

int azservo_min=AZSERVO_MIN, elservo_min=ELSERVO_MIN, azservo_max=AZSERVO_MAX, elservo_max=ELSERVO_MAX;

void setup() {
    //  pin and values are assigned to each instance  in driver but not exposed.
  azservo.attach(AZSERVO_PIN, azservo_min, azservo_max); // azimuth servo on pin 10, with min/max microseconds for 180 degree travel
  mywriteDegrees( azservo, 90, azservo_min, azservo_max );  // OEM "degrees" initializes halfway between 0/180 calibrations, Az = 90, East
  azservo.detach();
  
  elservo.attach(ELSERVO_PIN, elservo_min, elservo_max); // elevation servo on pin 9, with min/max microseconds for 180 degree travel
  mywriteDegrees( elservo, 90, elservo_min, elservo_max ); // OEM "degrees" initializes halfway between -90/90 calibrations, El = 0,Horizontal
  //elservo.detach();
  
  Serial.begin(115200);      // initialize Serial communication
  delay(10);    // wait a bit for Serial, in case it's present

  if (!IMU.begin()) {
    //Serial.println("Failed to initialize IMU!");
    while (true); // halt program
  }

  //Serial.print("IMU initialized, ");
  //Serial.print("Gyroscope sampling rate is "); //Serial.println(IMU.accelerationSampleRate());
  // check for the WiFi module:

  if (WiFi.status() == WL_NO_MODULE) {
    //Serial.println("Communication with WiFi module failed!");
    // don't continue
    while (true);
  }
  // Print firmware version on the module
  String fv = WiFi.firmwareVersion();
  Serial.print("Firmware version installed: "); Serial.println(fv);

  if (fv < WIFI_FIRMWARE_LATEST_VERSION) {
    //Serial.println("Please upgrade the firmware");
  }

  // attempt to connect to Wifi network:

  while (status != WL_CONNECTED) {
    //Serial.print("Attempting to connect to Network named: ");
    //Serial.println(ssid);                   // print the network name (SSID);
    // Connect to WPA/WPA2 network. Change this line if using open or WEP network:
    status = WiFi.begin(ssid, pass);
    // wait up to 10 seconds for connection:
    for (int timeout = 100; timeout > 0; timeout--) {
      if (status = WL_CONNECTED) {
        break;
      }
      delay(100);
    };
  }
  server.begin();                           // start the web server on port 80
  printWifiStatus();                        // you're connected now, so print out the status
}
/*
* **********************************************************************************************************************
*/
boolean  sendstatuspage = false;
void loop() { // spin here, looking for and servicing any client

  int i, microsecs;

  WiFiClient client = server.available();   // listen for incoming clients

  if (client) {                             // if we get a client then send a page and process any input
    IPAddress rip = client.remoteIP();
    //Serial.print(rip);
    //Serial.println(" connected\n"); // print a message out the //Serial port
    String currentLine = "";              // make Strings to hold incoming data from the client
    while (client.connected()) {            // loop while the client's connected & Bytes are available
      if (client.available()) {             // if there's bytes to read from the client,
        char c = client.read();             // read a byte, then
        // DEBUG
// Serial.write(c);                    //  print it out the //Serial monitor
        if (c == '\n') {

          if (currentLine.length() == 0) {
            // if the current line was blank, we have now received two newline characters in a row, the second on an empty line
            // I think this may be the default http end-of-connection signal in the absence of 'Content Length' or 'Chunked'.
            // so send an HTTP response page back to the client (we've already parsed any requests):

            // DEBUG
            //Serial.println("end of client request, now executing a response\n");
            if (sendstatuspage)
              sendStatusPage(client);
            else
              sendHomePage(client);
            break;

          } else     {  // currentLine was non-zero length & we got a 1st terminator so parse the line
            //build a char array so we can use indexOf to check for a GET request
            int str_len = currentLine.length() + 1;
            char char_array[str_len];
            currentLine.toCharArray(char_array, str_len);  // writes the String to a null terminated array of char

            if (currentLine.indexOf('GET') > 0) {
              String b = strtok(char_array, "?"); // first call to strtok returns the argument list following the '?'
              String variable = "NULL", value = "NULL";

              if (b != NULL) { // strip off one variable/value pair  "?variable = value"
                b = strtok(NULL, "=");  // get the variable before the " = ", (strtok is reentrant so hand it a NULL to continue)
                variable = b;
                value = strtok(NULL, " ");  // get the value, " " follows value and preceeds "HTTP"
 /*               
                Serial.println(""); // DEBUG
                Serial.print(variable); // DEBU
                Serial.print("="); // DEBUG
                Serial.println(value); // DEBUG
*/
                //   **** Now Execute any immediatee requests  (before returning any page)  ******
                // Az Rotor. increasing pulsewidth written to it result in CCW rotation of the antenna axis, looking 
                // from above down back toward the rotor. 
                // El Rotor has its drive shaft fixed and the body rotates. increasing pulsewidth written to the servo results
                // in CCW rotation of the antenna axis looking from the antenna towards the rotor

                // Considering the reference to be a vector pointing out the antenna mount axis, away from the Az/El assembly:
                // For user API (mywriteDegrees) calls, default setup of both rotors is mid-position with Az East=90 degrees and El H=0 degrees
                // For Az, user writing "0" should result in vector pointing North,  "180" should point due South
                // For El, the user writing "-90" results in the previously-north end of the horizontal dipole rotating down/CW 
                // writing "+90" also results in vertical polarization but with the prevously north end at the top,CCW rotation.
                //
                if (variable == "ElAngle") {
                  ElAngle = value.toFloat(); 
                  elservo.attach(ELSERVO_PIN);
                  // default mid position, user "0",  is horizontal, '90' from the servo hw's PoV, adjust by 90 degrees
                  mywriteDegrees( elservo, 90.0 + ElAngle - ElErrorAngle, elservo_min, elservo_max ); 
                  delay(250);
                  //elservo.detach();
                }
                // Mid-rotation is due East and 90 degrees
                //  looking down from above the az rotor, increasing angle/pulsewidth at the az rotor results in CW rotation
                //  so servo direction and value is reversed user PoV,  adjustment is necessary
                if (variable == "AzAngle") {
                  AzAngle = value.toFloat();
                  azservo.attach(AZSERVO_PIN);
                  mywriteDegrees( azservo, 180.0 - AzAngle, azservo_min, azservo_max ); 
                  
                  delay(250);
                  azservo.detach();
                }
                
                if (variable == "ElServoMin") {
                  elservo_min = value.toFloat(); 
                  elservo_min = constrain(elservo_min, SERVO_HARDLIMIT_LOW, SERVO_HARDLIMIT_HIGH);
                }
                 
                if (variable == "ElServoMax") {
                  elservo_max = value.toFloat(); 
                  elservo_max = constrain(elservo_max, SERVO_HARDLIMIT_LOW, SERVO_HARDLIMIT_HIGH);
                }
                 
                if (variable == "AzServoMin") {
                  azservo_min = value.toFloat(); 
                  azservo_min = constrain(azservo_min, SERVO_HARDLIMIT_LOW, SERVO_HARDLIMIT_HIGH);
                }
                 
                if (variable == "AzServoMax") {
                  azservo_max = value.toFloat(); 
                  azservo_max = constrain(azservo_max, SERVO_HARDLIMIT_LOW, SERVO_HARDLIMIT_HIGH);
                }
                
                if (variable == "Status") {
                  sendstatuspage = true;
                }

                if (variable == "Step") {
                  stepping = true;      // we will defer action until we send home page with steps included
                }
                if (variable == "Micro") {
                  micro_step =  value.toInt();      // with slowdown controls rotation rate
                }
                if (variable == "Slow") {
                  slowdown = value.toInt();      //  with micro_step controls rotation rate
                }
                if (variable == "Azusec") {
                  azservo.attach(AZSERVO_PIN);
                  MywriteMicroseconds(azservo, value.toInt());
                  delay(250);
                  azservo.detach();
                }
                if (variable == "Elusec") {
                  elservo.attach(ELSERVO_PIN);
                  MywriteMicroseconds(elservo, value.toInt());
                  delay(250);
                  //elservo.detach();
                }
              }  // end of parsing incoming request
            }  // end of a GET
            currentLine = "";
          }
        } // end last character was a \n terminator

        else if (c != '\r') {      // if we got anything else but a carriage return character (which is ignored),
          currentLine += c;      // keep building the currentLine
        }
      }     // end of bytes to read from the client,
    }      // end of connection
    // close the connection:
    client.stop();
    //Serial.println("client has been disconnected");
  }  // end found a client
}

void  MywriteMicroseconds(Servo servo, int microsecs) {
  // from previous position walk slowly up/down to the new value in micro_step size increments, waiting slowdown between each microstep
  prev_microsecs = servo.readMicroseconds();   // where the *driver* thinks the last write was, may have been forced from this during off time which will cause an initial jump until there is synch
  if (microsecs > prev_microsecs) {   //target position > present
    for (i = prev_microsecs; i < microsecs; i += micro_step)     {
      servo.writeMicroseconds(constrain(i,500,2500));
      delay(micro_step * slowdown);
      // a delay ~<slowdown> ms for every 1 us of requested angle change to allow Servo to slew and any gyro filter to settle
      //  if total useful range of the servo is 1500 +- 700 us then full range should take about 1.4 seconds at 1:1 scaling. This is still pretty fast
    }
  }
  if (microsecs < prev_microsecs) { //target position < present
    for (i = prev_microsecs; i > microsecs; i -= micro_step)     {
     
      servo.writeMicroseconds(constrain(i,500,2500));
      delay(micro_step * slowdown);
      // a delay ~1 ms for every 1 us of requested angle change to allow Servo to slew and any gyro filter to settle
      //  if total useful range of the servo is 1500 +- 700 us then full range should take about 1.4 seconds at 1:1 (slowdown=1)scaling.
      // increased by slowdown setting, e. g. 180 degree slew (delta of 1400 us) takes slighltly longer than (slowdown*1400/micro_step> ms
    }
  }
  delay(micro_step * slowdown);
  servo.writeMicroseconds(constrain(microsecs,500,2500));  // we've arrived, last bit that was less then micro_step so just go there
}

void  mywriteDegrees(Servo servo, float degree, int minpulse, int maxpulse) {   // NOTE!  This writes in terms of SERVO degrees, not user degrees
  int value;
  value = map(degree, 0, 180, minpulse, maxpulse);    // requested degrees turned into us for a particular servo
  MywriteMicroseconds(servo, value );
}

float cang, aX, aY, aZ, gX, gY, gZ;
float getIMU() {
  // Inside the El Rotor in the default condition, the Nano IMU is mounted so that Z points West, X points Down and Y point South
  while (!IMU.gyroscopeAvailable()) {};
  IMU.readGyroscope(gX, gY, gZ);
  while (!IMU.accelerationAvailable()) {};
  IMU.readAcceleration(aX, aY, aZ);
  cang = ((180/PI) * atan(aZ/aX));  // calculated angle, 
  //  positive aX is up-vertical
  //(z,x): (+,+)=> quadrant I 0 90       OK sign of Z is sign of angle
  //(z,x): (-,-)=> quadrant II 90 180
  //(z,x): (-,-)=> quadrant III -90 -180
  //(z,x): (+,-)=> quadrant IV -90 0      OK, sign of Z is sign of angle
  
  if(aX<0 && aZ<0 ) cang=cang-180;
  if(aX<0 && aZ>0 ) cang=cang+180;
  return (-cang);   // vertical is 0 but reverse sense, ccw is positive looking down axis toward rotator
}


void sendStatusPage( WiFiClient client ) {
  // HTTP headers always start with a response code (e.g. HTTP/1.1 200 OK)
  // and a content-type so the client knows what's coming, then a blank line:
  client.println("HTTP/1.1 200 OK");
  client.println("Content-type:text/html");
  client.println();

  client.print("<body style=\"background-color: lightgray\">");
  client.println("\n");client.println("<br>\n");
  
  client.println(" <b>   Voltages:</b>");client.println("<br>\n");
  int val;
  if ((val=analogRead(A0)) < 1023) {client.print("    V5+Kiwi = ");client.println(V0scale*LSB*val);client.println("<br>\n");}
  else { client.print("    V5+Kiwi = OV > ");client.print(V0scale*VCC); client.println(" VDC!<br>\n");}
  
  if ((val=analogRead(A1)) < 1023) {client.print("    V8+AzEl = ");client.println(V1scale*LSB*val);client.println("<br>\n");}
  else { client.print("    V8+AzEl = OV > ");client.print(V1scale*VCC); client.println(" VDC!<br>\n");}
  
  if ((val=analogRead(A2)) < 1023) {client.print("    V8+Preamp = ");client.println(V2scale*LSB*val);client.println("<br>\n");}
  else { client.print("    V8+Preamp = OV > ");client.print(V2scale*VCC); client.println(" VDC!<br>\n");}

  if ((val=analogRead(A3)) < 1023) {client.print("    V14+AA = ");client.println(V3scale*LSB*val);client.println("<br>\n");}
  else { client.print("    V14+AA = OV > ");client.print(V3scale*VCC); client.println(" VDC!<br>\n");}
  client.println("<br><br>\n");
  
  client.println(" <b>   Variables:</b>");client.println("<br>\n");
  client.print("    microstep size = ");client.println(String(micro_step));client.println("<br>\n");
  client.print("    slowdown = ");client.println(String(slowdown));client.println("<br><br>\n");
  client.print("    AzServoMin = "); client.println(String(azservo_min));client.println("<br>\n");
  client.print("    AzServoMax = "); client.println(String(azservo_max));client.println("<br>\n");
  client.print("    ElServoMin = "); client.println(String(elservo_min));client.println("<br>\n");
  client.print("    ElServoMax = "); client.println(String(elservo_max));client.println("<br><br><br>\n");

  client.println(" <b>   Settings:</b>");client.println("<br>\n");
  client.print("    AzAngle = "); client.println(String(AzAngle));client.println("<br>\n");
  client.print("    Az_us = "); client.println(String(azservo.readMicroseconds()));client.println("<br>\n");
  client.println("<br>\n");
  client.print("    ElAngle = "); client.println(String(ElAngle));client.println("<br>\n");
  client.print("    El_us = "); client.println(String(elservo.readMicroseconds()));client.println("<br>\n");
  client.print("    IMU says = "); client.println(String(getIMU()));client.println("<br>\n");
  client.println("<br><br>");
  client.println("    Accelerometer:");client.println("<br>\n");
  client.print("    aX="); client.println(String(aX));  client.print("     ");client.println("<br>\n");
  client.print("    aY="); client.println(String(aY));  client.print("     ");client.println("<br>\n");
  client.print("    aZ="); client.println(String(aZ));client.println("<br>\n");  
  client.print("<br>\n");
  client.print("    Gyroscope:");client.println("<br>\n"); 
  client.print("    gX="); client.println(String(gX));  client.print("     ");client.println("<br>\n");
  client.print("    gY="); client.println(String(gY));  client.print("     ");client.println("<br>\n");
  client.print("    gZ="); client.println(String(gZ));client.println("<br>\n");
  
  client.println("\n </body> ");
  // The HTTP response ends with another blank line:
  client.println();
  sendstatuspage = false;
}


void sendHomePage( WiFiClient client ) {
  // HTTP headers always start with a response code (e.g. HTTP/1.1 200 OK)
  // and a content-type so the client knows what's coming, then a blank line:
  client.println("HTTP/1.1 200 OK");
  client.println("Content-type:text/html");
  client.println();
  // the content of the HTTP response follows the header:
  client.print("<title>N6GN AzEl Rotor </title> ");
  client.print("<body style=\"background-color: lightgray\">");
  client.print(" <h1 style=\"color:black;font-size:40px;\">N6GN AzEl Rotor</h1>");
  client.print("<p style=\"color:blue;font-size:30px;\">Azimuth</p>");
  //
  client.println(" Azimuth goes from 0 degrees (North) through 90 degrees (DEFAULT East) to 180 degrees (South)");
  client.print("<br><br><form method=\"GET\">");
  client.print("<label form=\"quantity\">Select -10 < Azimuth Angle < 190  </label><br>");
  client.print("<input  style=\"background-color: lightblue\" type=\"number\" id=\"quantity\" name=\"AzAngle\" value=");
  client.print(String(AzAngle));
  client.print(" min=\"0\" max=\"180\">");
  client.println("</form>");
  client.print("<p style=\"color:green;font-size:30px;\"> <br>Elevation</p> ");


  client.println(" Elevation of 0 degrees (DEFAULT) is Up (Horizontal Polarization).  +-90 degrees are Vertical Polarizaton");
  client.print("<br><br>Requested ElAngle = ");
  client.print(String(ElAngle));
  client.print(" &  ");
  client.print("IMU Reads = ");
  client.print(String(getIMU()) );
  client.println("<br>");

  client.print("<form method=\"GET\">");
  client.print("<label form=\"quantity\">Select -100 < Elevation Angle < 100  </label><br>");
  client.println("<input style=\"background-color: lightgreen\" type=\"number\" id=\"quantity\" \nname=\"ElAngle\" value=");
  client.print(String(ElAngle));
  client.println("\n min=\"-100\" max=\"100\">");
  client.println("</form>");

  client.println("<br>\n Command line variable:value pair inputs are similar to the following: ");

  client.println("<br><br> \ncurl --get --data <a href=\"/?Status=1\"> \"Status=1\ http:\/\/ip:port\"</a> returns status information");
  client.println("<br><br>\nOther possible pairs:" );
  client.println("<br><br>\nAzAngle=xyz sets Azimuthal angle to xyz degrees" );
  client.println("<br>\nElAngle=xyz sets Elevation angle to xyz degrees" );
  client.print("<br>\nMicro=xyz sets \"chunks\" at ~7us/degree  together with " );
  client.print("Slow=xyz sets and scales rotation rate." );
  client.println("Full 180 degree range should take about 14 seconds at 1:1 (Slow=10)scaling." );

  client.println("<br> \nAzServoMin=xyz sets Azservo to xyz microseconds for 0 degrees ");
  client.println("<br> \nAzServoMax=xyz sets Azservo to xyz microseconds for 180 degrees ");
  client.println("<br> \nElServoMin=xyz sets Elservo to xyz microseconds for -90 degrees ");
  client.println("<br> \nElServoMax=xyz sets Elservo to xyz microseconds for 90 degrees ");
  client.println("<br> \nAzusec=xyz moves Azservo to xyz microseconds ");
  client.println("<br> \nElusec=xyz moves Elservo to xyz microseconds ");
  client.println("<br>");
  client.println("<br><br>\nTest Function:" );
  client.println("<br> \nStep=1 steps elevation over the entire range and compares with the IMU reading");
  client.println("<br>");   
  if (stepping) {
    float imu;
    elservo.attach(ELSERVO_PIN);
     ElAngle = -90;        // Start from default position
     mywriteDegrees(elservo, 90.0 + ElAngle - ElErrorAngle, elservo_min, elservo_max);
     delay(250);   //let IMU  settle
      client.println("<br>");
     for (ElAngle = -90; ElAngle <= 90; ElAngle += 5) {
      mywriteDegrees( elservo, 90.0 + ElAngle - ElErrorAngle, elservo_min, elservo_max ); // Have to keep track of servo parameters manually since driver object doesn't expose
      client.println("<br>");
      client.print("Requested ElAngle = ");
      client.print(String(ElAngle));
      client.print(",&nbsp;&nbsp;&nbsp;");
      
      delay(2000);   //let IMU  settle
      
      client.print("IMU Reads = ");
      client.print(String(imu=getIMU()) );

      client.print(",&nbsp;&nbsp;&nbsp;");
      client.print("Error = ");
      client.print(String(imu - ElAngle) );
    }
    
    client.println("<br>");
    ElAngle = 0;        // Return to default position when done
    mywriteDegrees(elservo, 90.0 + ElAngle - ElErrorAngle, elservo_min, elservo_max);
    stepping = false;

    //elservo.detach();
  }
  client.println("</body>");
  // The HTTP response ends with another blank line:
  client.println();
}


void printWifiStatus() {
  // print the SSID of the network you're attached to:
  //Serial.print("SSID: ");
  //Serial.println(WiFi.SSID());
  // print your board's IP address:
  IPAddress ip = WiFi.localIP();
  //Serial.print("Local IP Address: ");
  //Serial.println(ip);
  // print the received signal strength:
  long rssi = WiFi.RSSI();
  //Serial.print("signal strength (RSSI):");
  //Serial.print(rssi);
  //Serial.println(" dBm");
  // print where to go in a browser:
  //Serial.print("To see the web page in action, open a browser to http://");
  //Serial.print(ip);
  //Serial.print(":");

  //Serial.println(port);

  //Serial.println("");
  //Serial.println("Now waiting for a client connection");
}
